#ifndef PIXBUF_UTILITIES_H
#define PIXBUF_UTILITIES_H

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <arpa/inet.h>
#define DEFAULT_ALPHA 0xff

#define SET_IF_NULL(var, expr) if(var == NULL){ var = expr;}
#define SET_IF_NULL_AND_RETURN(var, expr) SET_IF_NULL(var, expr); return var;

GdkPixbuf* pixbuf_new_from_file(const gchar* rel_filename)
{
    gchar* filename = T2C_GET_DATA_PATH(rel_filename);
    GError* error = NULL;
    GdkPixbuf *result = gdk_pixbuf_new_from_file(filename, &error);
    free(filename);
    if(result == NULL)
    {
        TRACE("Cannot load pixbuf from file: %s.", error->message);
    }
    return result;
} 
GdkPixbufAnimation* animation_new_from_file(const gchar* rel_filename)
{
    gchar* filename = T2C_GET_DATA_PATH(rel_filename);
    GError* error = NULL;
    GdkPixbufAnimation *result = gdk_pixbuf_animation_new_from_file(filename, &error);
    free(filename);
    if(result == NULL)
    {
        TRACE("Cannot load animation from file: %s.", error->message);
    }
    return result;
} 



//structure of pixel in pixbuf
typedef struct
{
    guchar red, green, blue, alpha;
}Rgba;

void rgba_from_color(Rgba* dest, guint32 color)
{
    *(guint32*)dest = htonl(color);
}
guint32 rgba_to_color(Rgba* src)
{
    return ntohl(*(guint32*)src);
}
guint32 color_from_components(guchar r, guchar g, guchar b, guchar a)
{
    return ((guint32)r) << 24 | ((guint32)g) << 16 | ((guint32)b) << 8 | ((guint32)a);
//    Rgba rgba_color;
//    rgba_color.red = r;
//    rgba_color.green = g;
//    rgba_color.blue = b;
//    rgba_color.alpha = a;
//    return rgba_to_color(&rgba_color);
}
guint32 alpha_mask()
{
    return 0xff;
}
gboolean is_same_alpha(guint32 color1, guint32 color2)
{
    return (color1 & alpha_mask()) == (color2 & alpha_mask());
}
gboolean is_same_color_without_alpha(guint32 color1, guint32 color2)
{
    return (color1 & ~alpha_mask()) == (color2 & ~alpha_mask());
}
guint32 clear_alpha(guint32 color)
{
    return (color & ~alpha_mask()) | DEFAULT_ALPHA;
//    Rgba rgba_color;
//    rgba_from_color(&rgba_color, color);
//    rgba_color.alpha = DEFAULT_ALPHA;
//    return rgba_to_color(&rgba_color);
}
//convert color without alpha-channel to one, containing it
guint32 add_alpha(guint32 color)
{
    return color << 8 | DEFAULT_ALPHA;
}
//convert color with alpha-channel to one, doesn't containing it
guint32 remove_alpha(guint32 color)
{
    return color >> 8;
}
//gets value of pixel as guint32 according to has_alpha property
guint32 pixel_to_color(const gpointer pixel, gboolean has_alpha)
{
    Rgba rgba_result, *rgba_pixel = (Rgba*)pixel;
    if(has_alpha)
    {
        *(guint32*)(&rgba_result) = *(guint32*)(rgba_pixel);
    }
    else
    {
        rgba_result.red =   rgba_pixel->red;
        rgba_result.green = rgba_pixel->green;
        rgba_result.blue =  rgba_pixel->blue;
        rgba_result.alpha = DEFAULT_ALPHA;
    }
    return rgba_to_color(&rgba_result);
}
void set_pixel(const gpointer pixel, gboolean has_alpha, guint32 color)
{
    Rgba *rgba_pixel = (Rgba*)pixel;;
    if(!has_alpha)
    {
        Rgba rgba_color;
        rgba_from_color(&rgba_color, color);

        rgba_pixel->red =   rgba_color.red;
        rgba_pixel->green = rgba_color.green;
        rgba_pixel->blue =  rgba_color.blue;
    }
    else
    {
        rgba_from_color(rgba_pixel, color);
    }
}
//return pointer to the pixel
gpointer get_pixel(GdkPixbuf *pixbuf, int x, int y)
{
    int rowstride = gdk_pixbuf_get_rowstride(pixbuf);
    int n_channels = gdk_pixbuf_get_n_channels(pixbuf);
    guchar* pointer = gdk_pixbuf_get_pixels(pixbuf);
    return pointer + x * n_channels + y * rowstride;
}
guint32 get_pixel_value(GdkPixbuf *pixbuf, gint x, gint y)
{
    return pixel_to_color(get_pixel(pixbuf, x, y), gdk_pixbuf_get_has_alpha(pixbuf));
}

// This function tries to write every pixel of pixbuf
gboolean pixbuf_is_writable(GdkPixbuf *pixbuf)
{
    int n_channels = gdk_pixbuf_get_n_channels(pixbuf);
    int rowstride = gdk_pixbuf_get_rowstride(pixbuf);
    gboolean has_alpha = gdk_pixbuf_get_has_alpha(pixbuf);
    int width = gdk_pixbuf_get_width(pixbuf);
    int height = gdk_pixbuf_get_height(pixbuf); 
    gpointer pixels = gdk_pixbuf_get_pixels(pixbuf);
    for(int x = 0; x < width; x++)
    {
        for(int y = 0; y < height; y++)
        {
            guchar* pixel = (guchar*)pixels + x * n_channels + y * rowstride;
            set_pixel((gpointer)pixel, has_alpha, 0); 
        }
    }
    return TRUE;
}
//return true, if every pixel of pixbuf has its own address
gboolean pixbuf_is_full_writable(GdkPixbuf *pixbuf)
{
    int n_channels = gdk_pixbuf_get_n_channels(pixbuf);
    int rowstride = gdk_pixbuf_get_rowstride(pixbuf);
    int width = gdk_pixbuf_get_width(pixbuf);
    return rowstride >= width * n_channels;
}

//callback function for arbitrary generating data of pixels
typedef guint32 (*GetPixelValue) (int x, int y, gpointer user_data);
//Example of implementation of GetPixelValue; return the pixel of the pixbuf
typedef struct
{
    int rowstride;
    int n_channels;
    gpointer pixels;
    gboolean has_alpha;
}CallbackDataPixbuf;
void callback_data_pixbuf_load(CallbackDataPixbuf *data, GdkPixbuf* pixbuf)
{
    data->rowstride = gdk_pixbuf_get_rowstride(pixbuf);
    data->n_channels = gdk_pixbuf_get_n_channels(pixbuf);
    data->pixels = gdk_pixbuf_get_pixels(pixbuf);
    data->has_alpha = gdk_pixbuf_get_has_alpha(pixbuf);
}
guint32 get_pixel_of_pixbuf(int x, int y, CallbackDataPixbuf* data)
{
    guchar* pixel_pointer =
        (guchar*)data->pixels + x * data->n_channels + y * data->rowstride; 
    return pixel_to_color(pixel_pointer, data->has_alpha); 
}

//type of function which should determine if particular color of pixel is allowed
typedef gboolean (*PixelVerify) (int x, int y, guint32 pixel_color, gpointer user_data);

/*
 * Return TRUE if color of every pixel in pixbuf is expected(pixel_verify return TRUE).
 * Otherwise return FALSE and set x_diff and y_diff to coordinates of first incorrect point.
 */

gboolean pixbuf_content_verify(GdkPixbuf *pixbuf, gint *x_diff, gint *y_diff,
    PixelVerify pixel_verify, gpointer user_data)
{
    CallbackDataPixbuf data;
    callback_data_pixbuf_load(&data, pixbuf);
        
    int width = gdk_pixbuf_get_width(pixbuf);
    int height = gdk_pixbuf_get_height(pixbuf);
    
    for(int x = 0; x < width; x++)
    {
        for(int y = 0; y < height; y++)
        {
            if(!pixel_verify(x,y, get_pixel_of_pixbuf(x,y, &data), user_data))
            {
                *x_diff = x;
                *y_diff = y;
                return FALSE;
            }
        }
    }
    return TRUE;   
}
//function, which should return TRUE when pixel (%d,%d) should process
typedef gboolean (*PixelFilter)(int x, int y, gpointer data);

//example of filter - rectangular area
typedef struct
{
    int x,y;
    int width, height;
}CallbackDataPixbufArea;
//inside
gboolean filter_inside_area(int x, int y, CallbackDataPixbufArea* data)
{
    return (x >= data->x && 
            x < data->x + data->width &&
            y >= data->y &&
            y < data->y + data->height); 
}
//outside
gboolean filter_outside_area(int x, int y, CallbackDataPixbufArea* data)
{
    return !filter_inside_area(x, y, data);
}

/*
 * Same as pixbuf_content_verify,
 * but verify only filtered pixels(for which pixel_filter returns TRUE).
 */
 
gboolean pixbuf_content_verify_with_filter(GdkPixbuf *pixbuf, gint *x_diff, gint *y_diff,
    PixelFilter pixel_filter, gpointer filter_data,
    PixelVerify pixel_verify, gpointer user_data)
{
    CallbackDataPixbuf data;
    callback_data_pixbuf_load(&data, pixbuf);
        
    int width = gdk_pixbuf_get_width(pixbuf);
    int height = gdk_pixbuf_get_height(pixbuf);
    
    for(int x = 0; x < width; x++)
    {
        for(int y = 0; y < height; y++)
        {
            if(!pixel_filter(x, y, filter_data)) continue;
            
            if(!pixel_verify(x, y, get_pixel_of_pixbuf(x,y, &data), user_data))
            {
                *x_diff = x;
                *y_diff = y;
                return FALSE;
            }
        }
    }
    return TRUE;   
}
//
typedef struct
{
    GetPixelValue expected_pixel_value;
    gpointer user_data;
}CallbackDataVerifyExpected;
//"bridge" for verify if pixel value coincide with expected one
gboolean verify_expected(int x, int y, guint32 color, CallbackDataVerifyExpected *data)
{
    return color == data->expected_pixel_value(x, y, data->user_data);
}

/*
 * Return TRUE if color of every pixel in pixbuf coincides with color, returned by callback.
 * Otherwise return FALSE and set x_diff and y_diff to coordinates of first incorrect point.
 */

gboolean pixbuf_content_expected(GdkPixbuf *pixbuf, gint *x_diff, gint *y_diff,
    GetPixelValue expected_pixel_value, gpointer user_data)
{
    CallbackDataVerifyExpected data;
    data.expected_pixel_value = expected_pixel_value;
    data.user_data = user_data;
    
    return pixbuf_content_verify(pixbuf, x_diff, y_diff,
        (PixelVerify)verify_expected, &data); 
}
// Same function, but with filter.
gboolean pixbuf_content_expected_with_filter(GdkPixbuf *pixbuf, gint *x_diff, gint *y_diff,
    PixelFilter pixel_filter, gpointer filter_data,
    GetPixelValue expected_pixel_value, gpointer user_data)
{
    CallbackDataVerifyExpected data;
    data.expected_pixel_value = expected_pixel_value;
    data.user_data = user_data;
    
    return pixbuf_content_verify_with_filter(pixbuf, x_diff, y_diff,
        pixel_filter, filter_data,
        (PixelVerify)verify_expected, &data); 
}

//Types of difference in pixbufs
typedef enum
{
    DIFF_NO,
    DIFF_WIDTH,
    DIFF_HEIGHT,
    DIFF_ALPHA,
    DIFF_PIXEL_AT
}DiffType;
//Struct, describing information about difference
typedef struct
{
    DiffType type;
    //this fields have sence only when type is DIFF_PIXEL_AT
    int x,y;
}DiffInfo;

gboolean pixel_verify_without_alpha(int x, int y, guint32 color, CallbackDataPixbuf* data)
{
    return is_same_color_without_alpha(color, get_pixel_of_pixbuf(x, y, data));
}
gboolean verify_can_read(int x, int y, guint32 color)
{
    return TRUE;
}
/*
 * Compare two pixbufs. If them equals returns TRUE. Otherwise returns FALSE
 * and fills struct info, if it not NULL, with values, points to difference.
 */ 

gboolean pixbuf_equal(GdkPixbuf *pixbuf1, GdkPixbuf *pixbuf2,
                            gboolean compare_alpha, DiffInfo* info)
{
    DiffInfo info_tmp;
    DiffInfo *p_info = (info != NULL) ? info : &info_tmp;
    if(gdk_pixbuf_get_width(pixbuf1) != gdk_pixbuf_get_width(pixbuf2))
    {
        p_info->type = DIFF_WIDTH;
        return FALSE;
    }
    if(gdk_pixbuf_get_height(pixbuf1) != gdk_pixbuf_get_height(pixbuf2))
    {
        p_info->type = DIFF_HEIGHT;
        return FALSE;
    }
    if(compare_alpha)
    {
        if(gdk_pixbuf_get_has_alpha(pixbuf1) != gdk_pixbuf_get_has_alpha(pixbuf2))
        {
            p_info->type = DIFF_ALPHA;
            return FALSE;
        }
    }

    CallbackDataPixbuf data;
    callback_data_pixbuf_load(&data, pixbuf2);
    gboolean result;
    
    result = compare_alpha ?
                pixbuf_content_expected(pixbuf1, &p_info->x, &p_info->y,
                    (GetPixelValue)get_pixel_of_pixbuf, &data) :
                pixbuf_content_verify(pixbuf1, &p_info->x, &p_info->y,
                    (PixelVerify)is_same_color_without_alpha, &data);

    p_info->type = result ? DIFF_NO : DIFF_PIXEL_AT;
    return result;
}

//created pixbuf, filled by some data
GdkPixbuf* pixbuf_new_filled(int width, int height, gboolean has_alpha,
    GetPixelValue fill_func, gpointer user_data)
{
    GdkPixbuf *result = gdk_pixbuf_new(GDK_COLORSPACE_RGB, has_alpha, 8, width, height);
    int rowstride = gdk_pixbuf_get_rowstride(result);
    int n_channels = gdk_pixbuf_get_n_channels(result);
    guchar* pixels = gdk_pixbuf_get_pixels(result);
    for(int x = 0; x < width; x++)
    {
        for(int y = 0; y < height; y++)
        {
            guchar* pixel = pixels + x * n_channels + y * rowstride;
            set_pixel(pixel, has_alpha, fill_func(x, y, user_data));
        }
    }
    return result;
}
guint32 get_pixel_monochromatic(int x, int y, guint32* color)
{
    return *color;
}
GdkPixbuf* pixbuf_new_monochromatic(int width, int height, gboolean has_alpha, guint32 color)
{
    return pixbuf_new_filled(width, height, has_alpha,
        (GetPixelValue)get_pixel_monochromatic, &color);
}

#endif

