/*
 * This is a pixbuf-loader which loads images in "chb" (checkboard) format.
 */

#include <gdk-pixbuf/gdk-pixbuf.h>

#define GDK_PIXBUF_ENABLE_BACKEND
#include "gdk-pixbuf-io_internal.h"

#include <stdio.h>
#include <string.h>

typedef struct
{
    guchar r,g,b,a;
}Rgba;
gchar signature_string[] = {'C', 'h', 'e', 'c', 'k', 'b', 'o', 'a', 'r', 'd'};

/*
 * Format of a "chb" file:
 * struct
 * {
 *      gchar signature[] = {'C', 'h', 'e', 'c', 'k', 'b', 'o', 'a', 'r', 'd'};
 *      gint32 width, height;
 *      gint32 has_alpha;//0 - FALSE, 1 - TRUE
 *      Rgba rgba_color1, rgba_color2;
 * };
 */

#define SIZE_OFFSET (sizeof(signature_string))
#define COLORS_OFFSET (SIZE_OFFSET + 3 * sizeof(gint32))
#define BUFFER_SIZE (COLORS_OFFSET + 2 * sizeof(guint32))
//States of the loader(as of a Final State Machine)
typedef enum
{
    STATE_START,//not used
    STATE_READ_SIGNATURE,
    STATE_READ_SIZE,
    STATE_READ_COLORS,
    STATE_DONE,
    STATE_ERROR
}CheckboardLoaderState;
//Struct containing all information about loader
typedef struct
{
    GdkPixbufModuleSizeFunc size_func;
    GdkPixbufModulePreparedFunc prepare_func;
    GdkPixbufModuleUpdatedFunc update_func;
    gpointer user_data;
    
    int width,height;
    gboolean has_alpha;
    
    Rgba rgba_color1, rgba_color2;
    
    gsize bytes_read;
    CheckboardLoaderState state;
    
    gchar buffer[BUFFER_SIZE];
    
    GdkPixbuf* pixbuf;
}CheckboardLoaderContext;
//callback function(should initialize module context and return it)
gpointer checkboard_begin_load(   GdkPixbufModuleSizeFunc size_func,
                                    GdkPixbufModulePreparedFunc prepare_func,
                                    GdkPixbufModuleUpdatedFunc update_func,
                                    gpointer user_data,
                                    GError **error)
{
    CheckboardLoaderContext* context = g_malloc(sizeof(CheckboardLoaderContext));
    if(context == NULL)
    {
        g_set_error(error,
                    GDK_PIXBUF_ERROR,
                    GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
                    "Doesn't enough memory for allocate checkboard-loader context");
        return NULL;
    }
    context->size_func = size_func;
    context->prepare_func = prepare_func;
    context->update_func = update_func;
    context->user_data = user_data;
    
    context->bytes_read = 0;
    context->pixbuf = NULL;

    context->state = STATE_READ_SIGNATURE;
    
    return context;
}
gboolean checkboard_stop_load (gpointer context, GError **error)
{
    CheckboardLoaderContext* context_self = (CheckboardLoaderContext*) context; 
    //store state
    CheckboardLoaderState state = context_self->state;
    //free context if it need
    if(context_self != NULL)
    {
        if(context_self->pixbuf != NULL)
        {
            g_object_unref(context_self->pixbuf);
        }
        g_free(context_self);
    }

    if(state != STATE_DONE)
    {
        //image is not fully read
        g_set_error(error,
                        GDK_PIXBUF_ERROR,
                        GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                        (state == STATE_READ_SIGNATURE) ?
                        "Incorrect signature for checkboard format" :
                        "Unexpected end of stream while reading checkboard");
        return FALSE;
    }
    return TRUE;
}
//verify if the signature (or its part, if signature was not fully read) is correct
//If the signature was fully read, change state of the loader to STATE_READ_SIZE
gboolean read_signature(CheckboardLoaderContext* context, GError **error)
{
    if(strncmp(context->buffer, signature_string,
                MIN(sizeof(signature_string), context->bytes_read)) != 0)
    {
        g_set_error(error,
                        GDK_PIXBUF_ERROR,
                        GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                        "Incorrect signature for checkboard format");
        context->state = STATE_ERROR;
        return FALSE;
    }
    
    if(context->bytes_read >= SIZE_OFFSET)
    {
        context->state = STATE_READ_SIZE;
    }
    return TRUE;
}
//auxiliary functions
unsigned int read_uint32(gchar* stream)
{
    return ((guint32)stream[0]) << 24  | ((guint32)stream[1]) << 16
         | ((guint32)stream[2]) << 8 | ((guint32)stream[3]) << 0;
}
int read_int32(gchar* stream)
{
    return (int)read_uint32(stream);
}
void rgba_from_stream(Rgba* rgba, gchar* stream)
{
    rgba->r = (guchar)stream[0];
    rgba->g = (guchar)stream[1];
    rgba->b = (guchar)stream[2];
    rgba->a = (guchar)stream[3];
}
void rgba_write(Rgba* rgba_color, gboolean has_alpha, guchar* pixel)
{
    pixel[0] = rgba_color->r;
    pixel[1] = rgba_color->g;
    pixel[2] = rgba_color->b;
    if(has_alpha)
    {
        pixel[3] = rgba_color->a;
    }
}

void fill_checkboard(GdkPixbuf* pixbuf, gint check_size, Rgba* rgba_color1, Rgba* rgba_color2)
{
    int width = gdk_pixbuf_get_width(pixbuf);
    int height = gdk_pixbuf_get_width(pixbuf);
    guchar* pixels = gdk_pixbuf_get_pixels(pixbuf);
    int rowstride = gdk_pixbuf_get_rowstride(pixbuf);
    int n_channels = gdk_pixbuf_get_n_channels(pixbuf);
    gboolean has_alpha = gdk_pixbuf_get_has_alpha(pixbuf);
    
    int x,y;
    for(x = 0; x < width; x++)
    {
        for(y = 0; y < height; y++)
        {
            rgba_write((x / check_size + y / check_size) % 2 == 0 ? rgba_color1 : rgba_color2,
                        has_alpha, pixels + x * n_channels + y * rowstride);            
        }
    }
}

// Try to read information needed for creating pixbuf -
// size of pixbuf and transparency information.
// After this information is read fully, create pixbuf and
// change state of the loader to STATE_READ_COLORS.

gboolean read_size(CheckboardLoaderContext* context, GError **error)
{
    if(context->bytes_read >= COLORS_OFFSET)
    {
        context->width = read_int32(context->buffer + SIZE_OFFSET);
        context->height = read_int32(context->buffer + SIZE_OFFSET + sizeof(gint32));
        int alpha = read_int32(context->buffer + SIZE_OFFSET + 2 * sizeof(gint32));
        
        if(context->width <= 0 || context->height <= 0 || (alpha != 0 && alpha != 1))
        {
            g_set_error(error,
                        GDK_PIXBUF_ERROR,
                        GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                        "Information about size of checkboard corrupted");
            context->state = STATE_ERROR;
            return FALSE;
        }
        context->has_alpha = (alpha == 1);
        
        if(context->size_func != NULL)
        {
            int width = context->width;
            int height = context->height;
            
            context->size_func(&width, &height, context->user_data);
            
            if(width == 0 || height == 0)
            {
                context->state = STATE_DONE;
                return TRUE;
            }
        }
        
        context->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, context->has_alpha, 8,
                                            context->width, context->height);
        if(context->pixbuf == NULL)
        {
            printf("Cannot allocate pixbuf of size %d*%d.\n", context->width, context->height);
            g_set_error(error,
                        GDK_PIXBUF_ERROR,
                        GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
                        "Doesn't have enough memory for allocate pixbuf");
            context->state = STATE_ERROR;
            return FALSE;
        }
        if(context->prepare_func != NULL)
        {
            context->prepare_func(context->pixbuf, NULL, context->user_data);
        }
        context->state = STATE_READ_COLORS;
    }
    return TRUE;
}
// Read information about contents of the pixbuf.
// After that fill pixbuf with corresponds colors and change state of loader to STATE_DONE
gboolean read_colors(CheckboardLoaderContext* context, GError **error)
{
    if(context->bytes_read >= BUFFER_SIZE)
    {
        rgba_from_stream(&context->rgba_color1, context->buffer + COLORS_OFFSET);
        rgba_from_stream(&context->rgba_color2, context->buffer + COLORS_OFFSET  + sizeof(guint32));
        if(!context->has_alpha)
        {
            if(context->rgba_color1.a != 0 || context->rgba_color2.a != 0)
            {
                g_set_error(error,
                            GDK_PIXBUF_ERROR,
                            GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                            "Color has not-zero transparency, but pixbuf hasn't alpha-channel");
                context->state = STATE_ERROR;
                return FALSE;
            }
        }
        fill_checkboard(context->pixbuf, 4,
                            &context->rgba_color1, &context->rgba_color2);        
        
        if(context->update_func != NULL)
        {
            context->update_func(context->pixbuf,
                0,0,
                context->width, context->height,
                context->user_data);
        }
        context->state = STATE_DONE;
    }
    return TRUE;
}
//read stream data to buffer and run FSM
gboolean checkboard_load_increment (gpointer context, const guchar *buf,
                                    guint size, GError **error)
{
    CheckboardLoaderContext* context_self = (CheckboardLoaderContext*)context; 

    if(context_self->state == STATE_ERROR)
    {
        return FALSE;
    }
    
    if(context_self->state == STATE_DONE)
    {
        //ignore other data from stream;
        return TRUE;
    }
    
    gsize write_size = MIN(size, BUFFER_SIZE - context_self->bytes_read);
    memcpy(context_self->buffer + context_self->bytes_read, buf, write_size);

    context_self->bytes_read += write_size;

    //cannot use 'switch' because state may change inside functions
    if(context_self->state == STATE_READ_SIGNATURE)
    {
        if(!read_signature(context_self, error)) return FALSE;
    }
    
    if(context_self->state == STATE_READ_SIZE)
    {
        if(!read_size(context_self, error)) return FALSE;
    }
    
    if(context_self->state == STATE_READ_COLORS)
    {
        if(!read_colors(context_self, error)) return FALSE;
    }
    return TRUE;
}
//
void fill_info(GdkPixbufFormat* info)
{
    static GdkPixbufModulePattern signature[] = { 
        {signature_string, NULL, 100},
        {NULL, NULL, 0}};
    
    static gchar* mime_types[] = {NULL};//there is no mime type for our format :)
    
    static gchar* extensions[]= {"chb", NULL};
    
    
    info->name = "checkboard";
    info->signature = signature;
    
    info->description = "Format represented checkboard image";
    info->mime_types = mime_types;
    info->extensions = extensions;
    info->flags = 0;//not thread safe, cannot write, cannot scale
    info->disabled = FALSE;
    info->license = "other";
}
void fill_vtable(GdkPixbufModule* module)
{
    module->begin_load = checkboard_begin_load;
    module->stop_load = checkboard_stop_load;
    module->load_increment = checkboard_load_increment;
}
